package mjml

import (
	_ "embed" // embedd three js files
	"encoding/json"
	"log"
	"sync"

	"github.com/nayarsystems/go-duktape"
)

//go:embed js/dist/bundle.js
var bundle string

//go:embed js/dist/mjml-hide-config-error.js
var mjmlHCE string

// Error ...
type Error struct {
	Line             int    `json:"line"`
	Message          string `json:"message"`
	TagName          string `json:"tag_name"`
	FormattedMessage string `json:"formatted_message"`
}

func (err Error) Error() string {
	return err.FormattedMessage
}

// Errors ...
type Errors []Error

func (errors Errors) Error() string {
	s := ""
	if len(errors) > 0 {
		s = errors[0].Error()
	}

	if len(errors) > 1 {
		for _, err := range errors[1:] {
			s += "\n" + err.Error()
		}
	}

	return s
}

// Result ...
type Result struct {
	HTML    string `json:"html"`
	Errors  Errors `json:"errors"`
	MJML    string `json:"mjml"`
	Version string `json:"version"`
}

// Context is a duktape rendering context.
type Context struct {
	*duktape.Context
	m sync.Mutex
}

func initDukTapeCtx() (*Context, error) {
	ctx := duktape.New()
	if err := ctx.PevalString(mjmlHCE); err != nil {
		return nil, err
	}

	if err := ctx.PevalString(bundle); err != nil {
		return nil, err
	}

	if err := ctx.PevalString(`
	mjml
	`); err != nil {
		return nil, err
	}

	return &Context{Context: ctx}, nil
}

// ValidationLevel for mjml rendering.
type ValidationLevel string

// ValidationSkip your document is rendered without going through validation.
const ValidationSkip ValidationLevel = "skip"

// ValidationSoft your document is going through validation and is rendered, even if it has errors.
const ValidationSoft ValidationLevel = "soft"

// ValidationStrict your document is going through validation and is not rendered if it has any error.
const ValidationStrict ValidationLevel = "strict"

// Options for mjml rendering.
type Options struct {
	Beautify        bool
	Minify          bool
	Migrate         bool
	ValidationLevel ValidationLevel
}

func renderDukTape(ctx *Context, mjmlSrc string, options Options) (Result, error) {
	ctx.m.Lock()
	defer ctx.m.Unlock()

	stack := ctx.Context
	stack.Dup(-1)
	stack.RequireFunction(-1)
	stack.PushString(mjmlSrc)
	stack.PushObject()
	stack.PushNull()
	stack.PutPropString(-2, "filePath")

	if options.ValidationLevel != "" {
		stack.PushString(string(options.ValidationLevel))
		stack.PutPropString(-2, "validationLevel")
	}

	stack.PushBoolean(options.Minify)
	stack.PutPropString(-2, "minify")
	stack.PushBoolean(options.Migrate)
	stack.PutPropString(-2, "migrate")
	stack.Call(2)
	resultString := stack.GetString(-1)
	stack.Pop()

	var result Result
	if err := json.Unmarshal([]byte(resultString), &result); err != nil {
		log.Println(resultString)

		return result, err
	}

	return result, nil
}

/*
func initGoja() (*goja.Runtime, error) {
	vm := goja.New()
	b, err := ioutil.ReadFile("js/dist/mjml-hide-config-error.js")
	if err != nil {
		return nil, err
	}
	if _, err := vm.RunString(string(b)); err != nil {
		return nil, err
	}
	b, err = ioutil.ReadFile("js/dist/bundle.js")
	if err != nil {
		return nil, err
	}
	if _, err := vm.RunString(string(b)); err != nil {
		return nil, err
	}
	return vm, nil
}

func renderGoja(vm *goja.Runtime, mjmlSrc string) (string, error) {
	vm.Set("data", mjmlSrc)
	vm.RunString(`options={"beautify": true, "minify": false};`)
	v, err := vm.RunString("mjml(data, options);")
	if err != nil {
		return "", err
	}
	return v.Export().(string), nil
}

func main() {
	ctx := initDukTapeCtx()
	defer ctx.DestroyHeap()
	fmt.Println("duktape result is:", renderDukTape(ctx, mjmlSrc))
	fmt.Println("duktape result is:", renderDukTape(ctx, mjmlSrc))

	vm := initGoja()
	fmt.Println("goja result is:", renderGoja(vm, mjmlSrc))
	fmt.Println("goja result is:", renderGoja(vm, mjmlSrc))
}
*/
